@@ -16,9 +16,12 @@
 #include <libxml/xmlerror.h>
 #include <libxml/parserInternals.h>
 
+
+PG_MODULE_MAGIC;
+
 /* externally accessible functions */
 
-Datum		xml_valid(PG_FUNCTION_ARGS);
+Datum		xml_is_well_formed(PG_FUNCTION_ARGS);
 Datum		xml_encode_special_chars(PG_FUNCTION_ARGS);
 Datum		xpath_nodeset(PG_FUNCTION_ARGS);
 Datum		xpath_string(PG_FUNCTION_ARGS);
@@ -32,15 +35,6 @@ Datum		xpath_table(PG_FUNCTION_ARGS);
 void		elog_error(const char *explain, bool force);
 void		pgxml_parser_init(void);
 
-/* workspace for pgxml_xpath() */
-
-typedef struct
-{
-	xmlDocPtr	doctree;
-	xmlXPathContextPtr ctxt;
-	xmlXPathObjectPtr res;
-} xpath_workspace;
-
 /* local declarations */
 
 static void pgxml_errorHandler(void *ctxt, const char *msg,...);
@@ -54,10 +48,7 @@ static text *pgxml_result_to_text(xmlXPathObjectPtr res, xmlChar *toptag,
 
 static xmlChar *pgxml_texttoxmlchar(text *textstring);
 
-static xmlXPathObjectPtr pgxml_xpath(text *document, xmlChar *xpath,
-									 xpath_workspace *workspace);
-
-static void cleanup_workspace(xpath_workspace *workspace);
+static xmlXPathObjectPtr pgxml_xpath(text *document, xmlChar *xpath);
 
 /* Global variables */
 static char *pgxml_errorMsg = NULL;		/* overall error message */
@@ -149,10 +140,10 @@ pgxml_parser_init(void)
 
 /* Returns true if document is well-formed */
 
-PG_FUNCTION_INFO_V1(xml_valid);
+PG_FUNCTION_INFO_V1(xml_is_well_formed);
 
 Datum
-xml_valid(PG_FUNCTION_ARGS)
+xml_is_well_formed(PG_FUNCTION_ARGS)
 {
 	text	   *t = PG_GETARG_TEXT_P(0);		/* document buffer */
 	int32		docsize = VARSIZE(t) - VARHDRSZ;
@@ -301,22 +292,25 @@ PG_FUNCTION_INFO_V1(xpath_nodeset);
 Datum
 xpath_nodeset(PG_FUNCTION_ARGS)
 {
-	text	   *document = PG_GETARG_TEXT_P(0);
-	text	   *xpathsupp = PG_GETARG_TEXT_P(1);	/* XPath expression */
-	xmlChar    *toptag = pgxml_texttoxmlchar(PG_GETARG_TEXT_P(2));
-	xmlChar    *septag = pgxml_texttoxmlchar(PG_GETARG_TEXT_P(3));
-	xmlChar    *xpath;
-	text	   *xpres;
-	xmlXPathObjectPtr res;
-	xpath_workspace workspace;
+	xmlChar    *xpath,
+			   *toptag,
+			   *septag;
+	int32		pathsize;
+	text	   *xpathsupp,
+			   *xpres;
+
+	/* PG_GETARG_TEXT_P(0) is document buffer */
+	xpathsupp = PG_GETARG_TEXT_P(1);	/* XPath expression */
+
+	toptag = pgxml_texttoxmlchar(PG_GETARG_TEXT_P(2));
+	septag = pgxml_texttoxmlchar(PG_GETARG_TEXT_P(3));
+
+	pathsize = VARSIZE(xpathsupp) - VARHDRSZ;
 
 	xpath = pgxml_texttoxmlchar(xpathsupp);
 
-	res = pgxml_xpath(document, xpath, &workspace);
-
-	xpres = pgxml_result_to_text(res, toptag, septag, NULL);
-
-	cleanup_workspace(&workspace);
+	xpres = pgxml_result_to_text(pgxml_xpath(PG_GETARG_TEXT_P(0), xpath),
+								 toptag, septag, NULL);
 
 	pfree(xpath);
 
@@ -334,21 +328,23 @@ PG_FUNCTION_INFO_V1(xpath_list);
 Datum
 xpath_list(PG_FUNCTION_ARGS)
 {
-	text	   *document = PG_GETARG_TEXT_P(0);
-	text	   *xpathsupp = PG_GETARG_TEXT_P(1);	/* XPath expression */
-	xmlChar    *plainsep = pgxml_texttoxmlchar(PG_GETARG_TEXT_P(2));
-	xmlChar    *xpath;
-	text	   *xpres;
-	xmlXPathObjectPtr res;
-	xpath_workspace workspace;
+	xmlChar    *xpath,
+			   *plainsep;
+	int32		pathsize;
+	text	   *xpathsupp,
+			   *xpres;
+
+	/* PG_GETARG_TEXT_P(0) is document buffer */
+	xpathsupp = PG_GETARG_TEXT_P(1);	/* XPath expression */
+
+	plainsep = pgxml_texttoxmlchar(PG_GETARG_TEXT_P(2));
+
+	pathsize = VARSIZE(xpathsupp) - VARHDRSZ;
 
 	xpath = pgxml_texttoxmlchar(xpathsupp);
 
-	res = pgxml_xpath(document, xpath, &workspace);
-
-	xpres = pgxml_result_to_text(res, NULL, NULL, plainsep);
-
-	cleanup_workspace(&workspace);
+	xpres = pgxml_result_to_text(pgxml_xpath(PG_GETARG_TEXT_P(0), xpath),
+								 NULL, NULL, plainsep);
 
 	pfree(xpath);
 
@@ -363,13 +359,13 @@ PG_FUNCTION_INFO_V1(xpath_string);
 Datum
 xpath_string(PG_FUNCTION_ARGS)
 {
-	text	   *document = PG_GETARG_TEXT_P(0);
-	text	   *xpathsupp = PG_GETARG_TEXT_P(1);	/* XPath expression */
 	xmlChar    *xpath;
 	int32		pathsize;
-	text	   *xpres;
-	xmlXPathObjectPtr res;
-	xpath_workspace workspace;
+	text	   *xpathsupp,
+			   *xpres;
+
+	/* PG_GETARG_TEXT_P(0) is document buffer */
+	xpathsupp = PG_GETARG_TEXT_P(1);	/* XPath expression */
 
 	pathsize = VARSIZE(xpathsupp) - VARHDRSZ;
 
@@ -380,16 +376,13 @@ xpath_string(PG_FUNCTION_ARGS)
 	/* We could try casting to string using the libxml function? */
 
 	xpath = (xmlChar *) palloc(pathsize + 9);
-	strncpy((char *) xpath, "string(", 7);
 	memcpy((char *) (xpath + 7), VARDATA(xpathsupp), pathsize);
+	strncpy((char *) xpath, "string(", 7);
 	xpath[pathsize + 7] = ')';
 	xpath[pathsize + 8] = '\0';
 
-	res = pgxml_xpath(document, xpath, &workspace);
-
-	xpres = pgxml_result_to_text(res, NULL, NULL, NULL);
-
-	cleanup_workspace(&workspace);
+	xpres = pgxml_result_to_text(pgxml_xpath(PG_GETARG_TEXT_P(0), xpath),
+								 NULL, NULL, NULL);
 
 	pfree(xpath);
 
@@ -404,17 +397,21 @@ PG_FUNCTION_INFO_V1(xpath_number);
 Datum
 xpath_number(PG_FUNCTION_ARGS)
 {
-	text	   *document = PG_GETARG_TEXT_P(0);
-	text	   *xpathsupp = PG_GETARG_TEXT_P(1);	/* XPath expression */
 	xmlChar    *xpath;
+	int32		pathsize;
+	text	   *xpathsupp;
 	float4		fRes;
+
 	xmlXPathObjectPtr res;
-	xpath_workspace workspace;
+
+	/* PG_GETARG_TEXT_P(0) is document buffer */
+	xpathsupp = PG_GETARG_TEXT_P(1);	/* XPath expression */
+
+	pathsize = VARSIZE(xpathsupp) - VARHDRSZ;
 
 	xpath = pgxml_texttoxmlchar(xpathsupp);
 
-	res = pgxml_xpath(document, xpath, &workspace);
-
+	res = pgxml_xpath(PG_GETARG_TEXT_P(0), xpath);
 	pfree(xpath);
 
 	if (res == NULL)
@@ -422,8 +419,6 @@ xpath_number(PG_FUNCTION_ARGS)
 
 	fRes = xmlXPathCastToNumber(res);
 
-	cleanup_workspace(&workspace);
-
 	if (xmlXPathIsNaN(fRes))
 		PG_RETURN_NULL();
 
@@ -436,17 +431,21 @@ PG_FUNCTION_INFO_V1(xpath_bool);
 Datum
 xpath_bool(PG_FUNCTION_ARGS)
 {
-	text	   *document = PG_GETARG_TEXT_P(0);
-	text	   *xpathsupp = PG_GETARG_TEXT_P(1);	/* XPath expression */
 	xmlChar    *xpath;
+	int32		pathsize;
+	text	   *xpathsupp;
 	int			bRes;
+
 	xmlXPathObjectPtr res;
-	xpath_workspace workspace;
+
+	/* PG_GETARG_TEXT_P(0) is document buffer */
+	xpathsupp = PG_GETARG_TEXT_P(1);	/* XPath expression */
+
+	pathsize = VARSIZE(xpathsupp) - VARHDRSZ;
 
 	xpath = pgxml_texttoxmlchar(xpathsupp);
 
-	res = pgxml_xpath(document, xpath, &workspace);
-
+	res = pgxml_xpath(PG_GETARG_TEXT_P(0), xpath);
 	pfree(xpath);
 
 	if (res == NULL)
@@ -454,8 +453,6 @@ xpath_bool(PG_FUNCTION_ARGS)
 
 	bRes = xmlXPathCastToBoolean(res);
 
-	cleanup_workspace(&workspace);
-
 	PG_RETURN_BOOL(bRes);
 }
 
@@ -464,60 +461,48 @@ xpath_bool(PG_FUNCTION_ARGS)
 /* Core function to evaluate XPath query */
 
 static xmlXPathObjectPtr
-pgxml_xpath(text *document, xmlChar *xpath, xpath_workspace *workspace)
+pgxml_xpath(text *document, xmlChar *xpath)
 {
-	int32		docsize = VARSIZE(document) - VARHDRSZ;
+	xmlDocPtr	doctree;
+	xmlXPathContextPtr ctxt;
 	xmlXPathObjectPtr res;
 	xmlXPathCompExprPtr comppath;
+	int32		docsize;
 
-	workspace->doctree = NULL;
-	workspace->ctxt = NULL;
-	workspace->res = NULL;
+	docsize = VARSIZE(document) - VARHDRSZ;
 
 	pgxml_parser_init();
 
-	workspace->doctree = xmlParseMemory((char *) VARDATA(document), docsize);
-	if (workspace->doctree == NULL)
+	doctree = xmlParseMemory((char *) VARDATA(document), docsize);
+	if (doctree == NULL)
 		return NULL;			/* not well-formed */
 
-	workspace->ctxt = xmlXPathNewContext(workspace->doctree);
-	workspace->ctxt->node = xmlDocGetRootElement(workspace->doctree);
+	ctxt = xmlXPathNewContext(doctree);
+	ctxt->node = xmlDocGetRootElement(doctree);
 
 	/* compile the path */
 	comppath = xmlXPathCompile(xpath);
 	if (comppath == NULL)
 	{
-		cleanup_workspace(workspace);
+		xmlFreeDoc(doctree);
 		elog_error("XPath Syntax Error", true);
 	}
 
 	/* Now evaluate the path expression. */
-	res = xmlXPathCompiledEval(comppath, workspace->ctxt);
-	workspace->res = res;
-
+	res = xmlXPathCompiledEval(comppath, ctxt);
 	xmlXPathFreeCompExpr(comppath);
 
 	if (res == NULL)
-		cleanup_workspace(workspace);
+	{
+		xmlXPathFreeContext(ctxt);
+		xmlFreeDoc(doctree);
 
+		return NULL;
+	}
+	/* xmlFreeDoc(doctree); */
 	return res;
 }
 
-/* Clean up after processing the result of pgxml_xpath() */
-static void
-cleanup_workspace(xpath_workspace *workspace)
-{
-	if (workspace->res)
-		xmlXPathFreeObject(workspace->res);
-	workspace->res = NULL;
-	if (workspace->ctxt)
-		xmlXPathFreeContext(workspace->ctxt);
-	workspace->ctxt = NULL;
-	if (workspace->doctree)
-		xmlFreeDoc(workspace->doctree);
-	workspace->doctree = NULL;
-}
-
 static text *
 pgxml_result_to_text(xmlXPathObjectPtr res,
 					 xmlChar *toptag,
@@ -544,7 +529,7 @@ pgxml_result_to_text(xmlXPathObjectPtr res,
 			break;
 
 		default:
-			elog(NOTICE, "Unsupported XQuery result: %d", res->type);
+			elog(NOTICE, "unsupported XQuery result: %d", res->type);
 			xpresstr = xmlStrdup("<unsupported/>");
 	}
 
@@ -722,8 +707,8 @@ xpath_table(PG_FUNCTION_ARGS)
 	if (spi_tupdesc->natts != 2)
 	{
 		ereport(ERROR, (errcode(ERRCODE_INVALID_PARAMETER_VALUE),
-						errmsg("Expression returning multiple columns is not valid in parameter list"),
-						errdetail("Expected two columns in SPI result, got %d", spi_tupdesc->natts)));
+						errmsg("expression returning multiple columns is not valid in parameter list"),
+						errdetail("Expected two columns in SPI result, got %d.", spi_tupdesc->natts)));
 	}
 
 	/*
@@ -824,7 +809,7 @@ xpath_table(PG_FUNCTION_ARGS)
 								break;
 
 							default:
-								elog(NOTICE, "Unsupported XQuery result: %d", res->type);
+								elog(NOTICE, "unsupported XQuery result: %d", res->type);
 								resstr = xmlStrdup("<unsupported/>");
 						}
 
