@@ -3,12 +3,12 @@
  * joinpath.c
  *	  Routines to find all possible paths for processing a set of joins
  *
- * Portions Copyright (c) 1996-2003, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2006, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *
  * IDENTIFICATION
- *	  $Header: /cvsroot/pgsql/src/backend/optimizer/path/joinpath.c,v 1.82.2.3 2005/05/24 18:03:24 tgl Exp $
+ *	  $PostgreSQL: pgsql/src/backend/optimizer/path/joinpath.c,v 1.107.2.1 2007/05/22 01:40:42 tgl Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -16,25 +16,24 @@
 
 #include <math.h>
 
-#include "optimizer/clauses.h"
 #include "optimizer/cost.h"
 #include "optimizer/pathnode.h"
 #include "optimizer/paths.h"
-#include "parser/parsetree.h"
-#include "utils/lsyscache.h"
 
 
-static void sort_inner_and_outer(Query *root, RelOptInfo *joinrel,
+static void sort_inner_and_outer(PlannerInfo *root, RelOptInfo *joinrel,
 					 RelOptInfo *outerrel, RelOptInfo *innerrel,
 					 List *restrictlist, List *mergeclause_list,
 					 JoinType jointype);
-static void match_unsorted_outer(Query *root, RelOptInfo *joinrel,
+static void match_unsorted_outer(PlannerInfo *root, RelOptInfo *joinrel,
 					 RelOptInfo *outerrel, RelOptInfo *innerrel,
 					 List *restrictlist, List *mergeclause_list,
 					 JoinType jointype);
-static void hash_inner_and_outer(Query *root, RelOptInfo *joinrel,
+static void hash_inner_and_outer(PlannerInfo *root, RelOptInfo *joinrel,
 					 RelOptInfo *outerrel, RelOptInfo *innerrel,
 					 List *restrictlist, JoinType jointype);
+static Path *best_appendrel_indexscan(PlannerInfo *root, RelOptInfo *rel,
+						 RelOptInfo *outer_rel, JoinType jointype);
 static List *select_mergejoin_clauses(RelOptInfo *joinrel,
 						 RelOptInfo *outerrel,
 						 RelOptInfo *innerrel,
@@ -54,7 +53,7 @@ static List *select_mergejoin_clauses(RelOptInfo *joinrel,
  * paths found so far.
  */
 void
-add_paths_to_joinrel(Query *root,
+add_paths_to_joinrel(PlannerInfo *root,
 					 RelOptInfo *joinrel,
 					 RelOptInfo *outerrel,
 					 RelOptInfo *innerrel,
@@ -65,9 +64,9 @@ add_paths_to_joinrel(Query *root,
 
 	/*
 	 * Find potential mergejoin clauses.  We can skip this if we are not
-	 * interested in doing a mergejoin.  However, mergejoin is currently
-	 * our only way of implementing full outer joins, so override
-	 * mergejoin disable if it's a full join.
+	 * interested in doing a mergejoin.  However, mergejoin is currently our
+	 * only way of implementing full outer joins, so override mergejoin
+	 * disable if it's a full join.
 	 */
 	if (enable_mergejoin || jointype == JOIN_FULL)
 		mergeclause_list = select_mergejoin_clauses(joinrel,
@@ -95,23 +94,22 @@ add_paths_to_joinrel(Query *root,
 
 	/*
 	 * 3. Consider paths where the inner relation need not be explicitly
-	 * sorted.	This includes mergejoins only (nestloops were already
-	 * built in match_unsorted_outer).
+	 * sorted.	This includes mergejoins only (nestloops were already built in
+	 * match_unsorted_outer).
 	 *
 	 * Diked out as redundant 2/13/2000 -- tgl.  There isn't any really
-	 * significant difference between the inner and outer side of a
-	 * mergejoin, so match_unsorted_inner creates no paths that aren't
-	 * equivalent to those made by match_unsorted_outer when
-	 * add_paths_to_joinrel() is invoked with the two rels given in the
-	 * other order.
+	 * significant difference between the inner and outer side of a mergejoin,
+	 * so match_unsorted_inner creates no paths that aren't equivalent to
+	 * those made by match_unsorted_outer when add_paths_to_joinrel() is
+	 * invoked with the two rels given in the other order.
 	 */
 	match_unsorted_inner(root, joinrel, outerrel, innerrel,
 						 restrictlist, mergeclause_list, jointype);
 #endif
 
 	/*
-	 * 4. Consider paths where both outer and inner relations must be
-	 * hashed before being joined.
+	 * 4. Consider paths where both outer and inner relations must be hashed
+	 * before being joined.
 	 */
 	if (enable_hashjoin)
 		hash_inner_and_outer(root, joinrel, outerrel, innerrel,
@@ -133,7 +131,7 @@ add_paths_to_joinrel(Query *root,
  * 'jointype' is the type of join to do
  */
 static void
-sort_inner_and_outer(Query *root,
+sort_inner_and_outer(PlannerInfo *root,
 					 RelOptInfo *joinrel,
 					 RelOptInfo *outerrel,
 					 RelOptInfo *innerrel,
@@ -145,7 +143,7 @@ sort_inner_and_outer(Query *root,
 	Path	   *outer_path;
 	Path	   *inner_path;
 	List	   *all_pathkeys;
-	List	   *i;
+	ListCell   *l;
 
 	/*
 	 * If we are doing a right or full join, we must use *all* the
@@ -174,8 +172,8 @@ sort_inner_and_outer(Query *root,
 	/*
 	 * We only consider the cheapest-total-cost input paths, since we are
 	 * assuming here that a sort is required.  We will consider
-	 * cheapest-startup-cost input paths later, and only if they don't
-	 * need a sort.
+	 * cheapest-startup-cost input paths later, and only if they don't need a
+	 * sort.
 	 *
 	 * If unique-ification is requested, do it and then handle as a plain
 	 * inner join.
@@ -194,39 +192,37 @@ sort_inner_and_outer(Query *root,
 	}
 
 	/*
-	 * Each possible ordering of the available mergejoin clauses will
-	 * generate a differently-sorted result path at essentially the same
-	 * cost.  We have no basis for choosing one over another at this level
-	 * of joining, but some sort orders may be more useful than others for
-	 * higher-level mergejoins, so it's worth considering multiple
-	 * orderings.
+	 * Each possible ordering of the available mergejoin clauses will generate
+	 * a differently-sorted result path at essentially the same cost.  We have
+	 * no basis for choosing one over another at this level of joining, but
+	 * some sort orders may be more useful than others for higher-level
+	 * mergejoins, so it's worth considering multiple orderings.
 	 *
 	 * Actually, it's not quite true that every mergeclause ordering will
 	 * generate a different path order, because some of the clauses may be
-	 * redundant.  Therefore, what we do is convert the mergeclause list
-	 * to a list of canonical pathkeys, and then consider different
-	 * orderings of the pathkeys.
+	 * redundant.  Therefore, what we do is convert the mergeclause list to a
+	 * list of canonical pathkeys, and then consider different orderings of
+	 * the pathkeys.
 	 *
 	 * Generating a path for *every* permutation of the pathkeys doesn't seem
 	 * like a winning strategy; the cost in planning time is too high. For
-	 * now, we generate one path for each pathkey, listing that pathkey
-	 * first and the rest in random order.	This should allow at least a
-	 * one-clause mergejoin without re-sorting against any other possible
-	 * mergejoin partner path.	But if we've not guessed the right
-	 * ordering of secondary keys, we may end up evaluating clauses as
-	 * qpquals when they could have been done as mergeclauses. We need to
-	 * figure out a better way.  (Two possible approaches: look at all the
-	 * relevant index relations to suggest plausible sort orders, or make
-	 * just one output path and somehow mark it as having a sort-order
-	 * that can be rearranged freely.)
+	 * now, we generate one path for each pathkey, listing that pathkey first
+	 * and the rest in random order.  This should allow at least a one-clause
+	 * mergejoin without re-sorting against any other possible mergejoin
+	 * partner path.  But if we've not guessed the right ordering of secondary
+	 * keys, we may end up evaluating clauses as qpquals when they could have
+	 * been done as mergeclauses. We need to figure out a better way.  (Two
+	 * possible approaches: look at all the relevant index relations to
+	 * suggest plausible sort orders, or make just one output path and somehow
+	 * mark it as having a sort-order that can be rearranged freely.)
 	 */
 	all_pathkeys = make_pathkeys_for_mergeclauses(root,
 												  mergeclause_list,
 												  outerrel);
 
-	foreach(i, all_pathkeys)
+	foreach(l, all_pathkeys)
 	{
-		List	   *front_pathkey = lfirst(i);
+		List	   *front_pathkey = (List *) lfirst(l);
 		List	   *cur_pathkeys;
 		List	   *cur_mergeclauses;
 		List	   *outerkeys;
@@ -234,35 +230,34 @@ sort_inner_and_outer(Query *root,
 		List	   *merge_pathkeys;
 
 		/* Make a pathkey list with this guy first. */
-		if (i != all_pathkeys)
+		if (l != list_head(all_pathkeys))
 			cur_pathkeys = lcons(front_pathkey,
-								 lremove(front_pathkey,
-										 listCopy(all_pathkeys)));
+								 list_delete_ptr(list_copy(all_pathkeys),
+												 front_pathkey));
 		else
 			cur_pathkeys = all_pathkeys;		/* no work at first one... */
 
 		/*
 		 * Select mergeclause(s) that match this sort ordering.  If we had
-		 * redundant merge clauses then we will get a subset of the
-		 * original clause list.  There had better be some match,
-		 * however...
+		 * redundant merge clauses then we will get a subset of the original
+		 * clause list.  There had better be some match, however...
 		 */
 		cur_mergeclauses = find_mergeclauses_for_pathkeys(root,
 														  cur_pathkeys,
-													   mergeclause_list);
+														  mergeclause_list);
 		Assert(cur_mergeclauses != NIL);
 
 		/* Forget it if can't use all the clauses in right/full join */
 		if (useallclauses &&
-			length(cur_mergeclauses) != length(mergeclause_list))
+			list_length(cur_mergeclauses) != list_length(mergeclause_list))
 			continue;
 
 		/*
 		 * Build sort pathkeys for both sides.
 		 *
 		 * Note: it's possible that the cheapest paths will already be sorted
-		 * properly.  create_mergejoin_path will detect that case and
-		 * suppress an explicit sort step, so we needn't do so here.
+		 * properly.  create_mergejoin_path will detect that case and suppress
+		 * an explicit sort step, so we needn't do so here.
 		 */
 		outerkeys = make_pathkeys_for_mergeclauses(root,
 												   cur_mergeclauses,
@@ -299,10 +294,11 @@ sort_inner_and_outer(Query *root,
  *	  only outer paths that are already ordered well enough for merging).
  *
  * We always generate a nestloop path for each available outer path.
- * In fact we may generate as many as four: one on the cheapest-total-cost
+ * In fact we may generate as many as five: one on the cheapest-total-cost
  * inner path, one on the same with materialization, one on the
- * cheapest-startup-cost inner path (if different),
- * and one on the best inner-indexscan path (if any).
+ * cheapest-startup-cost inner path (if different), one on the
+ * cheapest-total inner-indexscan path (if any), and one on the
+ * cheapest-startup inner-indexscan path (if different).
  *
  * We also consider mergejoins if mergejoin clauses are available.	We have
  * two ways to generate the inner path for a mergejoin: sort the cheapest
@@ -324,7 +320,7 @@ sort_inner_and_outer(Query *root,
  * 'jointype' is the type of join to do
  */
 static void
-match_unsorted_outer(Query *root,
+match_unsorted_outer(PlannerInfo *root,
 					 RelOptInfo *joinrel,
 					 RelOptInfo *outerrel,
 					 RelOptInfo *innerrel,
@@ -338,15 +334,16 @@ match_unsorted_outer(Query *root,
 	Path	   *inner_cheapest_startup = innerrel->cheapest_startup_path;
 	Path	   *inner_cheapest_total = innerrel->cheapest_total_path;
 	Path	   *matpath = NULL;
-	Path	   *bestinnerjoin = NULL;
-	List	   *i;
+	Path	   *index_cheapest_startup = NULL;
+	Path	   *index_cheapest_total = NULL;
+	ListCell   *l;
 
 	/*
 	 * Nestloop only supports inner, left, and IN joins.  Also, if we are
-	 * doing a right or full join, we must use *all* the mergeclauses as
-	 * join clauses, else we will not have a valid plan.  (Although these
-	 * two flags are currently inverses, keep them separate for clarity
-	 * and possible future changes.)
+	 * doing a right or full join, we must use *all* the mergeclauses as join
+	 * clauses, else we will not have a valid plan.  (Although these two flags
+	 * are currently inverses, keep them separate for clarity and possible
+	 * future changes.)
 	 */
 	switch (jointype)
 	{
@@ -385,27 +382,37 @@ match_unsorted_outer(Query *root,
 	else if (nestjoinOK)
 	{
 		/*
-		 * If the cheapest inner path is a join or seqscan, we should
-		 * consider materializing it.  (This is a heuristic: we could
-		 * consider it always, but for inner indexscans it's probably a
-		 * waste of time.)
+		 * If the cheapest inner path is a join or seqscan, we should consider
+		 * materializing it.  (This is a heuristic: we could consider it
+		 * always, but for inner indexscans it's probably a waste of time.)
 		 */
 		if (!(IsA(inner_cheapest_total, IndexPath) ||
+			  IsA(inner_cheapest_total, BitmapHeapPath) ||
 			  IsA(inner_cheapest_total, TidPath)))
 			matpath = (Path *)
 				create_material_path(innerrel, inner_cheapest_total);
 
 		/*
-		 * Get the best innerjoin indexpath (if any) for this outer rel.
-		 * It's the same for all outer paths.
+		 * Get the best innerjoin indexpaths (if any) for this outer rel.
+		 * They're the same for all outer paths.
 		 */
-		bestinnerjoin = best_inner_indexscan(root, innerrel,
-											 outerrel->relids, jointype);
+		if (innerrel->reloptkind != RELOPT_JOINREL)
+		{
+			if (IsA(inner_cheapest_total, AppendPath))
+				index_cheapest_total = best_appendrel_indexscan(root,
+																innerrel,
+																outerrel,
+																jointype);
+			else if (innerrel->rtekind == RTE_RELATION)
+				best_inner_indexscan(root, innerrel, outerrel, jointype,
+									 &index_cheapest_startup,
+									 &index_cheapest_total);
+		}
 	}
 
-	foreach(i, outerrel->pathlist)
+	foreach(l, outerrel->pathlist)
 	{
-		Path	   *outerpath = (Path *) lfirst(i);
+		Path	   *outerpath = (Path *) lfirst(l);
 		List	   *merge_pathkeys;
 		List	   *mergeclauses;
 		List	   *innersortkeys;
@@ -416,8 +423,8 @@ match_unsorted_outer(Query *root,
 		int			sortkeycnt;
 
 		/*
-		 * If we need to unique-ify the outer path, it's pointless to
-		 * consider any but the cheapest outer.
+		 * If we need to unique-ify the outer path, it's pointless to consider
+		 * any but the cheapest outer.
 		 */
 		if (save_jointype == JOIN_UNIQUE_OUTER)
 		{
@@ -428,9 +435,9 @@ match_unsorted_outer(Query *root,
 		}
 
 		/*
-		 * The result will have this sort order (even if it is implemented
-		 * as a nestloop, and even if some of the mergeclauses are
-		 * implemented by qpquals rather than as true mergeclauses):
+		 * The result will have this sort order (even if it is implemented as
+		 * a nestloop, and even if some of the mergeclauses are implemented by
+		 * qpquals rather than as true mergeclauses):
 		 */
 		merge_pathkeys = build_join_pathkeys(root, joinrel, jointype,
 											 outerpath->pathkeys);
@@ -441,8 +448,8 @@ match_unsorted_outer(Query *root,
 			 * Always consider a nestloop join with this outer and
 			 * cheapest-total-cost inner.  When appropriate, also consider
 			 * using the materialized form of the cheapest inner, the
-			 * cheapest-startup-cost inner path, and the best innerjoin
-			 * indexpath.
+			 * cheapest-startup-cost inner path, and the cheapest innerjoin
+			 * indexpaths.
 			 */
 			add_path(joinrel, (Path *)
 					 create_nestloop_path(root,
@@ -470,13 +477,23 @@ match_unsorted_outer(Query *root,
 											  inner_cheapest_startup,
 											  restrictlist,
 											  merge_pathkeys));
-			if (bestinnerjoin != NULL)
+			if (index_cheapest_total != NULL)
 				add_path(joinrel, (Path *)
 						 create_nestloop_path(root,
 											  joinrel,
 											  jointype,
 											  outerpath,
-											  bestinnerjoin,
+											  index_cheapest_total,
+											  restrictlist,
+											  merge_pathkeys));
+			if (index_cheapest_startup != NULL &&
+				index_cheapest_startup != index_cheapest_total)
+				add_path(joinrel, (Path *)
+						 create_nestloop_path(root,
+											  joinrel,
+											  jointype,
+											  outerpath,
+											  index_cheapest_startup,
 											  restrictlist,
 											  merge_pathkeys));
 		}
@@ -493,8 +510,8 @@ match_unsorted_outer(Query *root,
 		/*
 		 * Done with this outer path if no chance for a mergejoin.
 		 *
-		 * Special corner case: for "x FULL JOIN y ON true", there will be
-		 * no join clauses at all.  Ordinarily we'd generate a clauseless
+		 * Special corner case: for "x FULL JOIN y ON true", there will be no
+		 * join clauses at all.  Ordinarily we'd generate a clauseless
 		 * nestloop path, but since mergejoin is our only join type that
 		 * supports FULL JOIN, it's necessary to generate a clauseless
 		 * mergejoin path instead.
@@ -502,11 +519,11 @@ match_unsorted_outer(Query *root,
 		if (mergeclauses == NIL)
 		{
 			if (jointype == JOIN_FULL)
-				/* okay to try for mergejoin */ ;
+				 /* okay to try for mergejoin */ ;
 			else
 				continue;
 		}
-		if (useallclauses && length(mergeclauses) != length(mergeclause_list))
+		if (useallclauses && list_length(mergeclauses) != list_length(mergeclause_list))
 			continue;
 
 		/* Compute the required ordering of the inner path */
@@ -515,9 +532,9 @@ match_unsorted_outer(Query *root,
 													   innerrel);
 
 		/*
-		 * Generate a mergejoin on the basis of sorting the cheapest
-		 * inner. Since a sort will be needed, only cheapest total cost
-		 * matters.  (But create_mergejoin_path will do the right thing if
+		 * Generate a mergejoin on the basis of sorting the cheapest inner.
+		 * Since a sort will be needed, only cheapest total cost matters. (But
+		 * create_mergejoin_path will do the right thing if
 		 * inner_cheapest_total is already correctly sorted.)
 		 */
 		add_path(joinrel, (Path *)
@@ -537,15 +554,18 @@ match_unsorted_outer(Query *root,
 			continue;
 
 		/*
-		 * Look for presorted inner paths that satisfy the innersortkey
-		 * list --- or any truncation thereof, if we are allowed to build
-		 * a mergejoin using a subset of the merge clauses.  Here, we
-		 * consider both cheap startup cost and cheap total cost.  Ignore
-		 * inner_cheapest_total, since we already made a path with it.
+		 * Look for presorted inner paths that satisfy the innersortkey list
+		 * --- or any truncation thereof, if we are allowed to build a
+		 * mergejoin using a subset of the merge clauses.  Here, we consider
+		 * both cheap startup cost and cheap total cost.  We can ignore
+		 * inner_cheapest_total on the first iteration, since we already made
+		 * a path with it --- but not on later iterations with shorter sort
+		 * keys, because then we are considering a different situation, viz
+		 * using a simpler mergejoin to avoid a sort of the inner rel.
 		 */
-		num_sortkeys = length(innersortkeys);
+		num_sortkeys = list_length(innersortkeys);
 		if (num_sortkeys > 1 && !useallclauses)
-			trialsortkeys = listCopy(innersortkeys);	/* need modifiable copy */
+			trialsortkeys = list_copy(innersortkeys);	/* need modifiable copy */
 		else
 			trialsortkeys = innersortkeys;		/* won't really truncate */
 		cheapest_startup_inner = NULL;
@@ -558,15 +578,16 @@ match_unsorted_outer(Query *root,
 
 			/*
 			 * Look for an inner path ordered well enough for the first
-			 * 'sortkeycnt' innersortkeys.	NB: trialsortkeys list is
-			 * modified destructively, which is why we made a copy...
+			 * 'sortkeycnt' innersortkeys.	NB: trialsortkeys list is modified
+			 * destructively, which is why we made a copy...
 			 */
-			trialsortkeys = ltruncate(sortkeycnt, trialsortkeys);
+			trialsortkeys = list_truncate(trialsortkeys, sortkeycnt);
 			innerpath = get_cheapest_path_for_pathkeys(innerrel->pathlist,
 													   trialsortkeys,
 													   TOTAL_COST);
 			if (innerpath != NULL &&
-				innerpath != inner_cheapest_total &&
+				(innerpath != inner_cheapest_total ||
+				 sortkeycnt < num_sortkeys) &&
 				(cheapest_total_inner == NULL ||
 				 compare_path_costs(innerpath, cheapest_total_inner,
 									TOTAL_COST) < 0))
@@ -601,7 +622,8 @@ match_unsorted_outer(Query *root,
 													   trialsortkeys,
 													   STARTUP_COST);
 			if (innerpath != NULL &&
-				innerpath != inner_cheapest_total &&
+				(innerpath != inner_cheapest_total ||
+				 sortkeycnt < num_sortkeys) &&
 				(cheapest_startup_inner == NULL ||
 				 compare_path_costs(innerpath, cheapest_startup_inner,
 									STARTUP_COST) < 0))
@@ -610,8 +632,8 @@ match_unsorted_outer(Query *root,
 				if (innerpath != cheapest_total_inner)
 				{
 					/*
-					 * Avoid rebuilding clause list if we already made
-					 * one; saves memory in big join trees...
+					 * Avoid rebuilding clause list if we already made one;
+					 * saves memory in big join trees...
 					 */
 					if (newclauses == NIL)
 					{
@@ -619,8 +641,8 @@ match_unsorted_outer(Query *root,
 						{
 							newclauses =
 								find_mergeclauses_for_pathkeys(root,
-														   trialsortkeys,
-														   mergeclauses);
+															   trialsortkeys,
+															   mergeclauses);
 							Assert(newclauses != NIL);
 						}
 						else
@@ -663,7 +685,7 @@ match_unsorted_outer(Query *root,
  * 'jointype' is the type of join to do
  */
 static void
-hash_inner_and_outer(Query *root,
+hash_inner_and_outer(PlannerInfo *root,
 					 RelOptInfo *joinrel,
 					 RelOptInfo *outerrel,
 					 RelOptInfo *innerrel,
@@ -672,7 +694,7 @@ hash_inner_and_outer(Query *root,
 {
 	bool		isouterjoin;
 	List	   *hashclauses;
-	List	   *i;
+	ListCell   *l;
 
 	/*
 	 * Hashjoin only supports inner, left, and IN joins.
@@ -696,15 +718,15 @@ hash_inner_and_outer(Query *root,
 	 * We need to build only one hashpath for any given pair of outer and
 	 * inner relations; all of the hashable clauses will be used as keys.
 	 *
-	 * Scan the join's restrictinfo list to find hashjoinable clauses that
-	 * are usable with this pair of sub-relations.
+	 * Scan the join's restrictinfo list to find hashjoinable clauses that are
+	 * usable with this pair of sub-relations.
 	 */
 	hashclauses = NIL;
-	foreach(i, restrictlist)
+	foreach(l, restrictlist)
 	{
-		RestrictInfo *restrictinfo = (RestrictInfo *) lfirst(i);
+		RestrictInfo *restrictinfo = (RestrictInfo *) lfirst(l);
 
-		if (restrictinfo->left_relids == NULL ||
+		if (!restrictinfo->can_join ||
 			restrictinfo->hashjoinoperator == InvalidOid)
 			continue;			/* not hashjoinable */
 
@@ -712,7 +734,7 @@ hash_inner_and_outer(Query *root,
 		 * If processing an outer join, only use its own join clauses for
 		 * hashing.  For inner joins we need not be so picky.
 		 */
-		if (isouterjoin && restrictinfo->ispusheddown)
+		if (isouterjoin && restrictinfo->is_pushed_down)
 			continue;
 
 		/*
@@ -724,7 +746,7 @@ hash_inner_and_outer(Query *root,
 			/* righthand side is inner */
 		}
 		else if (bms_is_subset(restrictinfo->left_relids, innerrel->relids) &&
-			 bms_is_subset(restrictinfo->right_relids, outerrel->relids))
+				 bms_is_subset(restrictinfo->right_relids, outerrel->relids))
 		{
 			/* lefthand side is inner */
 		}
@@ -738,9 +760,9 @@ hash_inner_and_outer(Query *root,
 	if (hashclauses)
 	{
 		/*
-		 * We consider both the cheapest-total-cost and
-		 * cheapest-startup-cost outer paths.  There's no need to consider
-		 * any but the cheapest-total-cost inner path, however.
+		 * We consider both the cheapest-total-cost and cheapest-startup-cost
+		 * outer paths.  There's no need to consider any but the
+		 * cheapest-total-cost inner path, however.
 		 */
 		Path	   *cheapest_startup_outer = outerrel->cheapest_startup_path;
 		Path	   *cheapest_total_outer = outerrel->cheapest_total_path;
@@ -781,6 +803,77 @@ hash_inner_and_outer(Query *root,
 	}
 }
 
+/*
+ * best_appendrel_indexscan
+ *	  Finds the best available set of inner indexscans for a nestloop join
+ *	  with the given append relation on the inside and the given outer_rel
+ *	  outside.	Returns an AppendPath comprising the best inner scans, or
+ *	  NULL if there are no possible inner indexscans.
+ *
+ * Note that we currently consider only cheapest-total-cost.  It's not
+ * very clear what cheapest-startup-cost might mean for an AppendPath.
+ */
+static Path *
+best_appendrel_indexscan(PlannerInfo *root, RelOptInfo *rel,
+						 RelOptInfo *outer_rel, JoinType jointype)
+{
+	int			parentRTindex = rel->relid;
+	List	   *append_paths = NIL;
+	bool		found_indexscan = false;
+	ListCell   *l;
+
+	foreach(l, root->append_rel_list)
+	{
+		AppendRelInfo *appinfo = (AppendRelInfo *) lfirst(l);
+		int			childRTindex;
+		RelOptInfo *childrel;
+		Path	   *index_cheapest_startup;
+		Path	   *index_cheapest_total;
+
+		/* append_rel_list contains all append rels; ignore others */
+		if (appinfo->parent_relid != parentRTindex)
+			continue;
+
+		childRTindex = appinfo->child_relid;
+		childrel = find_base_rel(root, childRTindex);
+		Assert(childrel->reloptkind == RELOPT_OTHER_MEMBER_REL);
+
+		/*
+		 * Check to see if child was rejected by constraint exclusion. If so,
+		 * it will have a cheapest_total_path that's an Append path with no
+		 * members (see set_plain_rel_pathlist).
+		 */
+		if (IsA(childrel->cheapest_total_path, AppendPath) &&
+			((AppendPath *) childrel->cheapest_total_path)->subpaths == NIL)
+			continue;			/* OK, we can ignore it */
+
+		/*
+		 * Get the best innerjoin indexpaths (if any) for this child rel.
+		 */
+		best_inner_indexscan(root, childrel, outer_rel, jointype,
+							 &index_cheapest_startup, &index_cheapest_total);
+
+		/*
+		 * If no luck on an indexpath for this rel, we'll still consider an
+		 * Append substituting the cheapest-total inner path.  However we must
+		 * find at least one indexpath, else there's not going to be any
+		 * improvement over the base path for the appendrel.
+		 */
+		if (index_cheapest_total)
+			found_indexscan = true;
+		else
+			index_cheapest_total = childrel->cheapest_total_path;
+
+		append_paths = lappend(append_paths, index_cheapest_total);
+	}
+
+	if (!found_indexscan)
+		return NULL;
+
+	/* Form and return the completed Append path. */
+	return (Path *) create_append_path(rel, append_paths);
+}
+
 /*
  * select_mergejoin_clauses
  *	  Select mergejoin clauses that are usable for a particular join.
@@ -799,55 +892,33 @@ select_mergejoin_clauses(RelOptInfo *joinrel,
 {
 	List	   *result_list = NIL;
 	bool		isouterjoin = IS_OUTER_JOIN(jointype);
-	List	   *i;
+	bool		have_nonmergeable_joinclause = false;
+	ListCell   *l;
 
-	foreach(i, restrictlist)
+	foreach(l, restrictlist)
 	{
-		RestrictInfo *restrictinfo = (RestrictInfo *) lfirst(i);
+		RestrictInfo *restrictinfo = (RestrictInfo *) lfirst(l);
 
 		/*
-		 * If processing an outer join, only use its own join clauses in
-		 * the merge.  For inner joins we need not be so picky.
-		 *
-		 * Furthermore, if it is a right/full join then *all* the explicit
-		 * join clauses must be mergejoinable, else the executor will
-		 * fail. If we are asked for a right join then just return NIL to
-		 * indicate no mergejoin is possible (we can handle it as a left
-		 * join instead). If we are asked for a full join then emit an
-		 * error, because there is no fallback.
+		 * If processing an outer join, only use its own join clauses in the
+		 * merge.  For inner joins we can use pushed-down clauses too. (Note:
+		 * we don't set have_nonmergeable_joinclause here because pushed-down
+		 * clauses will become otherquals not joinquals.)
 		 */
-		if (isouterjoin)
-		{
-			if (restrictinfo->ispusheddown)
-				continue;
-			switch (jointype)
-			{
-				case JOIN_RIGHT:
-					if (restrictinfo->left_relids == NULL ||
-						restrictinfo->mergejoinoperator == InvalidOid)
-						return NIL;		/* not mergejoinable */
-					break;
-				case JOIN_FULL:
-					if (restrictinfo->left_relids == NULL ||
-						restrictinfo->mergejoinoperator == InvalidOid)
-						ereport(ERROR,
-								(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
-								 errmsg("FULL JOIN is only supported with merge-joinable join conditions")));
-					break;
-				default:
-					/* otherwise, it's OK to have nonmergeable join quals */
-					break;
-			}
-		}
+		if (isouterjoin && restrictinfo->is_pushed_down)
+			continue;
 
-		if (restrictinfo->left_relids == NULL ||
+		if (!restrictinfo->can_join ||
 			restrictinfo->mergejoinoperator == InvalidOid)
+		{
+			have_nonmergeable_joinclause = true;
 			continue;			/* not mergejoinable */
+		}
 
 		/*
 		 * Check if clause is usable with these input rels.  All the vars
-		 * needed on each side of the clause must be available from one or
-		 * the other of the input rels.
+		 * needed on each side of the clause must be available from one or the
+		 * other of the input rels.
 		 */
 		if (bms_is_subset(restrictinfo->left_relids, outerrel->relids) &&
 			bms_is_subset(restrictinfo->right_relids, innerrel->relids))
@@ -855,15 +926,42 @@ select_mergejoin_clauses(RelOptInfo *joinrel,
 			/* righthand side is inner */
 		}
 		else if (bms_is_subset(restrictinfo->left_relids, innerrel->relids) &&
-			 bms_is_subset(restrictinfo->right_relids, outerrel->relids))
+				 bms_is_subset(restrictinfo->right_relids, outerrel->relids))
 		{
 			/* lefthand side is inner */
 		}
 		else
+		{
+			have_nonmergeable_joinclause = true;
 			continue;			/* no good for these input relations */
+		}
 
 		result_list = lcons(restrictinfo, result_list);
 	}
 
+	/*
+	 * If it is a right/full join then *all* the explicit join clauses must be
+	 * mergejoinable, else the executor will fail. If we are asked for a right
+	 * join then just return NIL to indicate no mergejoin is possible (we can
+	 * handle it as a left join instead). If we are asked for a full join then
+	 * emit an error, because there is no fallback.
+	 */
+	if (have_nonmergeable_joinclause)
+	{
+		switch (jointype)
+		{
+			case JOIN_RIGHT:
+				return NIL;		/* not mergejoinable */
+			case JOIN_FULL:
+				ereport(ERROR,
+						(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
+						 errmsg("FULL JOIN is only supported with merge-joinable join conditions")));
+				break;
+			default:
+				/* otherwise, it's OK to have nonmergeable join quals */
+				break;
+		}
+	}
+
 	return result_list;
 }
